#!/bin/bash
# Purpose: Create a Linux template from cloud-image files with cloud-init support

# Script configuration variables
TEST=false  # Set to true to enable test mode (commands will be echoed but not executed)
KEEP=false  # Set to true to keep downloaded files after script execution
SKIP=false  # Set to true to skip certain steps (useful for debugging)

# Color variables for output formatting
green='\033[0;32m'  # Green color for success messages
red='\033[0;31m'    # Red color for error messages
clear='\033[0m'     # Clear color formatting

function run_as_root () {
  if [ "$EUID" -ne 0 ]; then
    echo -e "${red}Please run this script as root. Use 'sudo ./create_template.sh'${clear}"
    exit 1
  fi
}

function linux_selection () {
  echo ''
  echo "Create a Linux template based on..."
  local index=1
  local distros=()
  for config_file in distro_configs/*.conf; do
    local distro_name=$(basename "$config_file" .conf)
    distros+=("$distro_name")
    echo "  - [$index]  $distro_name"
    ((index++))
  done

  if [ ${#distros[@]} -eq 0 ]; then
    echo -e "${red}No Linux distribution configuration files found in 'distro_configs' directory.${clear}"
    exit 1
  fi

  while true; do
    read -p 'Select the number corresponding to the Linux distribution: ' selection
    if [[ $selection -ge 1 && $selection -le ${#distros[@]} ]]; then
      LINUX=${distros[$((selection-1))]}
      break
    else
      echo "Invalid selection. Please enter a number between 1 and ${#distros[@]}."
    fi
  done
  echo "Selected distribution: $LINUX"
}

function get_available_storage () {
  echo ''
  echo "Retrieving available storage options..."
  if ! command -v pvesh &> /dev/null; then
    echo -e "${red}Error: pvesh command not found. Please ensure Proxmox VE API tools are installed.${clear}"
    exit 1
  fi

  STORAGE_OPTIONS=$(pvesh get /storage --output-format json | jq -r '.[] | select(.content | contains("images")) | "\(.storage) (\(.type))"')
  if [ -z "$STORAGE_OPTIONS" ]; then
    echo -e "${red}No storage options found that support images. Please check your Proxmox VE setup.${clear}"
    exit 1
  fi
  echo "Available storage options:"
  echo "$STORAGE_OPTIONS"
}

function storage_selection () {
  get_available_storage
  while true; do
    echo ''
    echo "Create the template on this storage..."
    echo "$STORAGE_OPTIONS" | nl -w 2 -s '. '
    read -p 'Select the storage number: ' selection
    STORAGE=$(echo "$STORAGE_OPTIONS" | sed -n "${selection}p" | awk '{print $1}')
    if [ -n "$STORAGE" ]; then
      break
    else
      echo "Invalid selection. Please enter a number corresponding to the storage option."
    fi
  done
  echo "Selected storage: $STORAGE"
}

function network_selection () {
  local default_bridge="vmbr0"
  while true; do
    echo ''
    echo "Which network bridge should be used..."
    echo "Network bridge vmbr0 "
    echo "Network bridge vmbr1 "
    read -p "Choose network bridge (default is $default_bridge): " selection
    case $selection in
      * ) BRIDGE=vmbr0; break;;
      * ) BRIDGE=vmbr1; break;;
      * )    BRIDGE=$default_bridge; break;;
    esac
  done
  echo "Selected network bridge: $BRIDGE"
}

# Not ready for production, shows more bridges than I expected
#network_selection() {
#    # List all network bridges
#    bridges=$(brctl show | awk 'NR>1 {print $1}' | sort | uniq)
#
#    # Check if any bridges were found
#    if [ -z "$bridges" ]; then
#        echo "No network bridges found."
#        return 1
#    fi
#
#    echo "Select a network bridge:"
#    select bridge in $bridges; do
#        if [ -n "$bridge" ]; then
#            echo "You selected: $bridge"
#            break
#        else
#            echo "Invalid selection. Please try again."
#        fi
#    done
#}

function keyboard_selection () {
  local default_layout="sv"
  while true; do
    echo ''
    echo "Which keyboard layout to use."
    echo "  -  sv    (default)"
    echo "  -  en-us"
    echo "  -  en-gb"
    read -p "Select keyboard layout (default is $default_layout): " selection
    case "$selection" in
      * ) KEYBOARD=sv;     break;;
      * ) KEYBOARD=en-us;  break;;
      * ) KEYBOARD=en-gb;  break;;
      * ) KEYBOARD=$default_layout; break;;
    esac
  done
  echo "Selected keyboard layout: $KEYBOARD"
}

function pool_selection () {
  local default_pool="tmpl"
  local pools

  echo ''
  echo "Retrieving available pools..."
  pools=$(pvesh get /pools --output-format json | jq -r '.[].poolid')

  if ! echo "$pools" | grep -q "^$default_pool$"; then
    read -p "The pool '$default_pool' does not exist. Do you want to create it? [Y/n]: " create_pool
    case $create_pool in
      [Yy]* | "" ) pvesh create /pools -poolid $default_pool
                   echo "Pool '$default_pool' created.";;
      [Nn]* ) echo "Proceeding without creating the pool.";;
      * ) echo "Invalid input. Proceeding without creating the pool.";;
    esac
  fi

  # Retrieve available pools again to update the list
  pools=$(pvesh get /pools --output-format json | jq -r '.[].poolid')

  if [ -n "$pools" ]; then
    while true; do
      echo ''
      echo "Do you want to add the VM/template to a pool?"
      echo "$pools" | nl -w 2 -s '. '
      echo "  -  None"
      read -p 'Select the pool number (or 0 for none): ' selection
      if [[ $selection -eq 0 ]]; then
        POOL=
        break
      elif [[ $selection -ge 1 && $selection -le $(echo "$pools" | wc -l) ]]; then
        POOL="--pool $(echo "$pools" | sed -n "${selection}p")"
        break
      else
        echo "Invalid selection. Please enter a number corresponding to the pool option."
      fi
    done
  else
    POOL=
  fi
  echo "Selected pool: $POOL"
}

function load_distro_config () {
  local distro_name=$1
  local config_file="distro_configs/${distro_name}.conf"

  if [ -f "$config_file" ]; then
    source "$config_file"
  else
    echo -e "${red}Error: Configuration file for ${distro_name} not found in 'distro_configs' directory.${clear}"
    exit 1
  fi
}

function install_requirement () {
  if ! command -v virt-customize &> /dev/null; then
    echo ''
    echo "The package 'libguestfs-tools' is required but not installed."
    echo '###########################################################################'
    if [ $TEST = true ]; then
      echo "Running in test mode. Commands to be executed:"
      echo "Command: apt update"
      echo "Command: apt install libguestfs-tools -y"
    else
      echo "Updating package list and installing 'libguestfs-tools'..."
      apt update > /dev/null
      if apt install libguestfs-tools -y > /dev/null; then
        printf " ${green}\xE2\x9C\x94${clear} Package 'libguestfs-tools' installed successfully.\n"
      else
        echo -e "${red}Failed to install 'libguestfs-tools'. Please install it manually and rerun the script.${clear}"
        exit 1
      fi
    fi
  fi

  if ! command -v jq &> /dev/null; then
    echo ''
    echo "The package 'jq' is required but not installed."
    echo '###########################################################################'
    if [ $TEST = true ]; then
      echo "Running in test mode. Commands to be executed:"
      echo "Command: apt update"
      echo "Command: apt install jq -y"
    else
      echo "Updating package list and installing 'jq'..."
      apt update > /dev/null
      if apt install jq -y > /dev/null; then
        printf " ${green}\xE2\x9C\x94${clear} Package 'jq' installed successfully.\n"
      else
        echo -e "${red}Failed to install 'jq'. Please install it manually and rerun the script.${clear}"
        exit 1
      fi
    fi
  fi
}

function start_time () {
  START_TIME=$(date +'%F %T %Z')
  echo ''
  echo ''
  echo "Script started at: $START_TIME"
}

function stop_time () {
  STOP_TIME=$(date +'%F %T %Z')
  echo ''
  echo "Template has been created!"
  echo "Script completed at: $STOP_TIME"
  echo ''
}

function write_variables () {
  echo ''
  echo "Writing out variables"
  echo '############################################################################'
  echo "USER=$USER"
  echo "OPTION=$OPTION"
  echo "LINUX=$LINUX"
  echo "NUM=$NUM"
  echo "STORAGE=$STORAGE"
  echo "VMID=$VMID"
  echo "DESC=$DESC"
  echo "FILE=$FILE"
  echo "EXT=$EXT"
  echo "KB=$KB"
  echo "KEYBOARD=$KEYBOARD"
  echo "URL=$URL"
  echo "INSTALL=$INSTALL"
  echo "FIRSTBOOT_INSTALL=$FIRSTBOOT_INSTALL"
  echo "UNINSTALL=$UNINSTALL"
  echo "FIRSTBOOT_COMMAND=$FIRSTBOOT_COMMAND"
  echo "FIRSTBOOT_SCRIPT=$FIRSTBOOT_SCRIPT"
  echo "TEST=$TEST"
  echo "KEEP=$KEEP"
  echo "SKIP=$SKIP"
  echo "START_TIME=$START_TIME"
  echo "STOP_TIME=$STOP_TIME"
  echo "POOL=$POOL"
  echo "virt-customize --add $FILE $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_SCRIPT"
  echo '############################################################################'
}

function temporary_image () {
  echo ''
  echo "Copying '$LINUX-downloaded.$EXT' to a temporary image file"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: cp $LINUX-downloaded.$EXT $LINUX-modified.$EXT"
    cp $LINUX-downloaded.$EXT $LINUX-modified.$EXT
  else
    cp $LINUX-downloaded.$EXT $LINUX-modified.$EXT > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf " ${green}\xE2\x9C\x94${clear} $LINUX-modified.$EXT has been saved.\n"
  else
    echo -e "${red}Failed to copy $LINUX-downloaded.$EXT to $LINUX-modified.$EXT.${clear}"
    exit 1
  fi
}

function customize_image () {
  echo ''
  echo "Customizing '$LINUX-modified.$EXT' image file"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: virt-customize --add $LINUX-modified.$EXT $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_COMMAND $FIRSTBOOT_SCRIPT"
    virt-customize --add $LINUX-modified.$EXT $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_COMMAND $FIRSTBOOT_SCRIPT
  else
    virt-customize --add $LINUX-modified.$EXT $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_COMMAND $FIRSTBOOT_SCRIPT > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf " ${green}\xE2\x9C\x94${clear} The '$LINUX-modified.$EXT' image file has been customized.\n"
  else
    echo -e "${red}Failed to customize $LINUX-modified.$EXT.${clear}"
    exit 1
  fi
}

function sysprep () {
  echo ''
  echo "Sysprepping the '$LINUX-modified.$EXT' image file"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: virt-sysprep --add $LINUX-modified.$EXT"
    virt-sysprep --add $LINUX-modified.$EXT
  else
    virt-sysprep --add $LINUX-modified.$EXT > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf " ${green}\xE2\x9C\x94${clear} The '$LINUX-modified.$EXT' image file has been sysprepped.\n"
  else
    echo -e "${red}Failed to sysprep $LINUX-modified.$EXT.${clear}"
    exit 1
  fi
}

function remove_machine_id () {
  echo ''
  echo "Removing machine-id from the '$LINUX-modified.$EXT' image file"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: virt-customize --add $LINUX-modified.$EXT --truncate /etc/machine-id"
    virt-customize --add $LINUX-modified.$EXT --truncate /etc/machine-id
  else
    virt-customize --add $LINUX-modified.$EXT --truncate /etc/machine-id > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf " ${green}\xE2\x9C\x94${clear} Machine-id has been removed from the '$LINUX-modified.$EXT' image file.\n"
  else
    echo -e "${red}Failed to remove machine-id from $LINUX-modified.$EXT.${clear}"
    exit 1
  fi
}

function create_vm () {
  echo ''
  echo "Creating Virtual Machine"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo ''
    echo "Command: qm create $VMID --name $LINUX-cloudinit-template --memory $MEMORY --cores $CORES --sockets $SOCKETS --cpu $CPU --net0 virtio,bridge=$BRIDGE $POOL"
    qm create $VMID --name $LINUX-cloudinit-template --memory $MEMORY --cores $CORES --sockets $SOCKETS --cpu $CPU --net0 virtio,bridge=$BRIDGE $POOL
  else
    qm create $VMID --name $LINUX-cloudinit-template --memory $MEMORY --cores $CORES --sockets $SOCKETS --cpu $CPU --net0 virtio,bridge=$BRIDGE $POOL > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} VM created\n"
    printf "       * Template: $LINUX-cloudinit-template\n"
    printf "       * VMID:     $VMID\n"
    printf "       * CPU type: $CPU\n"
    printf "       * Sockets:  $SOCKETS\n"
    printf "       * Cores:    $CORES\n"
    printf "       * Memory:   $MEMORY\n"
    printf "       * Bridge:   $BRIDGE\n"
    printf "       * Pool:     $POOL\n"
  else
    echo -e "${red}Failed to create VM.${clear}"
    exit 1
  fi
}

function import_disk () {
  echo ''
  echo "Importing disk image to VM"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: qm disk import $VMID $LINUX-modified.$EXT $STORAGE"
    qm disk import $VMID $LINUX-modified.$EXT $STORAGE
  else
    qm disk import $VMID $LINUX-modified.$EXT $STORAGE > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Disk image imported\n"
    printf "       * Storage:  $STORAGE\n"
  else
    echo -e "${red}Failed to import disk image to VM.${clear}"
    exit 1
  fi
}

function system_disk () {
  echo ''
  echo "Configuring system disk on VM"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: qm set $VMID --scsihw virtio-scsi-pci --scsi0 $STORAGE:vm-$VMID-disk-0,discard=on,ssd=1"
    qm set $VMID --scsihw virtio-scsi-pci --scsi0 $STORAGE:vm-$VMID-disk-0,discard=on,ssd=1
  else
    qm set $VMID --scsihw virtio-scsi-pci --scsi0 $STORAGE:vm-$VMID-disk-0,discard=on,ssd=1 > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Configured boot disk\n"
  else
    echo -e "${red}Failed to configure boot disk.${clear}"
    exit 1
  fi

  echo ''
  echo "Setting system disk as boot disk"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: qm set $VMID --boot c --bootdisk scsi0"
    qm set $VMID --boot c --bootdisk scsi0
  else
    qm set $VMID --boot c --bootdisk scsi0 > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Configured system disk to boot disk\n"
  else
    echo -e "${red}Failed to set system disk as boot disk.${clear}"
    exit 1
  fi
}

function create_cloud_init () {
  echo ''
  echo "Creating cloud-init drive on VM"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: qm set $VMID --ide2 $STORAGE:cloudinit"
    qm set $VMID --ide2 $STORAGE:cloudinit
  else
    qm set $VMID --ide2 $STORAGE:cloudinit > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Cloud-init drive created\n"
  else
    echo -e "${red}Failed to create cloud-init drive.${clear}"
    exit 1
  fi
}

function serial_vga () {
  echo ''
  echo "Configuring serial VGA on VM"
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo "Command: qm set $VMID --serial0 socket --vga serial0"
    qm set $VMID --serial0 socket --vga serial0
  else
    qm set $VMID --serial0 socket --vga serial0 > /dev/null
  fi

  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Serial VGA configured\n"
  else
    echo -e "${red}Failed to configure serial VGA.${clear}"
    exit 1
  fi
}

function enable_agent() {
  echo ''
  echo "Enabling agent on VM"
  echo '############################################################################'

  # Command to enable agent
  local command="qm set $VMID --agent enabled=1"

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Agent enabled\n"
  else
    echo -e "${red}Failed to enable agent.${clear}"
    exit 1
  fi
}

function disable_ciupgrade() {
  echo ''
  echo "Disabling Cloud-init Upgrade on VM"
  echo '############################################################################'

  # Command to disable Cloud-init Upgrade
  local command="qm set $VMID --ciupgrade 0"

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Cloud-init Upgrade disabled\n"
  else
    echo -e "${red}Failed to disable Cloud-init Upgrade.${clear}"
    exit 1
  fi
}

function configure_network() {
  echo ''
  echo "Configuring network on VM"
  echo '############################################################################'

  # Command to configure network with DHCP
  local command="qm set $VMID --ipconfig0 ip=dhcp"

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} DHCP configured\n"
  else
    echo -e "${red}Failed to configure DHCP.${clear}"
    exit 1
  fi
}

function configure_keyboard() {
  echo ''
  echo "Configuring keyboard on VM"
  echo '############################################################################'

  # Command to configure keyboard
  local command="qm set $VMID --keyboard $KEYBOARD"

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Keyboard configured\n"
    printf "       * Keyboard: $KEYBOARD\n"
  else
    echo -e "${red}Failed to configure keyboard.${clear}"
    exit 1
  fi
}

function add_description() {
  echo ''
  echo "Adding description to VM"
  echo '############################################################################'

  # Command to add description
  local command="qm set $VMID --description \"Server: <br> OS: $DESC <br> Created: $START_TIME\""

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Added Description\n"
    printf "       * OS:       $DESC\n"
    printf "       * Created:  $START_TIME\n"
  else
    echo -e "${red}Failed to add description.${clear}"
    exit 1
  fi
}

function add_tags() {
  echo ''
  echo "Adding tags to VM"
  echo '############################################################################'

  # Define tags
  local TAGS="template,$LINUX"

  # Command to add tags
  local command="qm set $VMID --tags \"$TAGS\""

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Added Tags\n"
    printf "       * Tags:     $TAGS\n"
  else
    echo -e "${red}Failed to add tags.${clear}"
    exit 1
  fi
}

function convert_to_template() {
  echo ''
  echo "Converting VM to template"
  echo '############################################################################'

  # Command to convert VM to template
  local command="qm template $VMID"

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf "   ${green}\xE2\x9C\x94${clear} Converted to template\n"
  else
    echo -e "${red}Failed to convert to template.${clear}"
    exit 1
  fi
}

function download_source() {
  echo ''
  echo "Download source image file"
  echo '############################################################################'

  # Command to download the source image file
  local command="wget $URL/$FILE.$EXT --progress=bar --show-progress -O $LINUX-downloaded.$EXT"

  if [ ! -f "$LINUX-downloaded.$EXT" ]; then
    if [ "$TEST" = true ]; then
      echo "Command: $command"
      $command
    else
      $command > /dev/null
    fi

    # Check the exit status of the command
    if [ $? -eq 0 ]; then
      printf " ${green}\xE2\x9C\x94${clear} The file '$FILE.$EXT' downloaded.\n"
    else
      echo -e "${red}Failed to download the file '$FILE.$EXT'.${clear}"
      exit 1
    fi
  else
    if [ "$KEEP" = true ]; then
      printf " \xE2\x9C\x94 The file download skipped (KEEP=true).\n"
    else
      if [ "$TEST" = true ]; then
        echo "Command: $command"
        $command
      else
        $command > /dev/null
      fi

      # Check the exit status of the command
      if [ $? -eq 0 ]; then
        printf " ${green}\xE2\x9C\x94${clear} The file '$FILE.$EXT' downloaded.\n"
      else
        echo -e "${red}Failed to download the file '$FILE.$EXT'.${clear}"
        exit 1
      fi
    fi
  fi
}

function remove_source() {
  echo ''
  echo "Removing source image files"
  echo '############################################################################'

  # Function to remove a file if it exists
  remove_file() {
    local file=$1
    local description=$2

    if [ -f "$file" ]; then
      echo ''
      echo "Remove $description image file"
      echo '############################################################################'

      # Command to remove the file
      local command="rm $file"

      if [ "$TEST" = true ]; then
        echo "Command: $command"
        $command
      else
        $command > /dev/null
      fi

      # Check the exit status of the command
      if [ $? -eq 0 ]; then
        printf " ${green}\xE2\x9C\x94${clear} '$file' image file removed.\n"
      else
        echo -e "${red}Failed to remove '$file' image file.${clear}"
        exit 1
      fi
    fi
  }

  # Remove modified image file
  remove_file "$LINUX-modified.$EXT" "modified"

  # Remove downloaded image file if KEEP is false
  if [ "$KEEP" = true ]; then
    printf " \xE2\x9C\x94 Keeping '$LINUX-downloaded.$EXT' image file.\n"
  else
    remove_file "$LINUX-downloaded.$EXT" "downloaded"
  fi
}

function destroy_vm() {
  echo ''
  echo "Purging Template $VMID from related configurations."
  echo '############################################################################'

  # Command to destroy the VM
  local command="qm destroy $VMID --destroy-unreferenced-disks 1 --purge 1"

  if [ "$TEST" = true ]; then
    echo "Command: $command"
    $command
  else
    $command > /dev/null
  fi

  # Check the exit status of the command
  if [ $? -eq 0 ]; then
    printf " ${green}\xE2\x9C\x94${clear} Template $VMID purged.\n"
  else
    echo -e "${red}Failed to purge Template $VMID.${clear}"
    exit 1
  fi
}

function main() {
  echo "Starting VM template creation process"
  echo '############################################################################'

  # Ensure the script is run as root
  run_as_root

  # Selection steps
  linux_selection
  storage_selection
  network_selection
  keyboard_selection
  pool_selection

  # Load distribution configuration
  load_distro_config "$LINUX"

  # Write variables if in test mode
  if [ "$TEST" = true ]; then
    write_variables
  fi

  # Exit if SKIP is true
  if [ "$SKIP" = true ]; then
    echo "Exiting..."
    exit 0
  fi

  # Start time tracking
  start_time

  # Execute steps with error handling
  install_requirement || { echo "Failed to install requirements"; exit 1; }
  remove_source || { echo "Failed to remove source"; exit 1; }
  destroy_vm || { echo "Failed to destroy VM"; exit 1; }
  download_source || { echo "Failed to download source"; exit 1; }
  temporary_image || { echo "Failed to create temporary image"; exit 1; }
  customize_image || { echo "Failed to customize image"; exit 1; }
  sysprep || { echo "Failed to sysprep"; exit 1; }
  remove_machine_id || { echo "Failed to remove machine ID"; exit 1; }
  create_vm || { echo "Failed to create VM"; exit 1; }
  import_disk || { echo "Failed to import disk"; exit 1; }
  system_disk || { echo "Failed to configure system disk"; exit 1; }
  create_cloud_init || { echo "Failed to create cloud-init"; exit 1; }
  serial_vga || { echo "Failed to configure serial VGA"; exit 1; }
  enable_agent || { echo "Failed to enable agent"; exit 1; }
  disable_ciupgrade || { echo "Failed to disable cloud-init upgrade"; exit 1; }
  configure_network || { echo "Failed to configure network"; exit 1; }
  configure_keyboard || { echo "Failed to configure keyboard"; exit 1; }
  add_description || { echo "Failed to add description"; exit 1; }
  add_tags || { echo "Failed to add tags"; exit 1; }
  convert_to_template || { echo "Failed to convert to template"; exit 1; }
  remove_source || { echo "Failed to remove source"; exit 1; }

  # Stop time tracking
  stop_time

  echo "VM template creation process completed successfully"
  exit 0
}

main
