#!/bin/bash
# Purpose = Create a Linux template from cloud-image files with cloud-init support

### Variables for running this script, for more information see the README-md file.
  TEST=false
  KEEP=false
  SKIP=false

### Variables for Create VM, these variables doesn't have to be changed. It's better to set this when cloning the template to a new virtual machine.
  MEMORY=1024
  CORES=1
  SOCKETS=1
  CPU='host'

### Set color variables, for color output
green='\033[0;32m'
red='\033[0;31m'
clear='\033[0m'

function ubuntu_focal () {  ### Ubuntu Focal
  VMID=2004
  DESC='Ubuntu 20.04 LTS (Focal Fossa)'
  FILE=focal-server-cloudimg-amd64
  EXT=img
  URL=https://cloud-images.ubuntu.com/focal/current
  INSTALL="--install qemu-guest-agent"
  FIRSTBOOT_INSTALL="--firstboot-install ansible,colordiff,git,haveged,mc,needrestart,tmux,qemu-guest-agent"
  UNINSTALL="--uninstall byobu"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_ubuntu-focal.sh"
}

function ubuntu_jammy () {  ### Ubuntu Jammy
  VMID=2204
  DESC='Ubuntu 22.04 LTS (Jammy Jellyfish)'
  FILE=jammy-server-cloudimg-amd64
  EXT=img
  URL=https://cloud-images.ubuntu.com/jammy/current
  INSTALL="--install qemu-guest-agent"
  FIRSTBOOT_INSTALL="--firstboot-install ansible,colordiff,git,haveged,mc,needrestart,tmux,qemu-guest-agent"
  UNINSTALL="--uninstall byobu"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_ubuntu-jammy.sh"
}

function alma_8 () {  ### AlmaLinux 8
  VMID=1888
  DESC='AlmaLinux 8 (Green Obsidian)'
  FILE=AlmaLinux-8-GenericCloud-latest.x86_64
  EXT=qcow2
  URL=https://repo.almalinux.org/almalinux/8/cloud/x86_64/images
  INSTALL="--install mc,tmux,qemu-guest-agent"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_alma-8.sh"
}

function alma_9 () {  ### AlmaLinux 9
  VMID=1999
  DESC='AlmaLinux 9 (Blue Onyx)'
  FILE=AlmaLinux-9-GenericCloud-latest.x86_64
  EXT=qcow2
  URL=https://repo.almalinux.org/almalinux/9/cloud/x86_64/images
  INSTALL="--install mc,tmux,qemu-guest-agent"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_alma-9.sh"
}

function rocky_8 () {  ### Rocky 8
  VMID=2888
  DESC='Rocky Linux 8 (Green Obsidian)'
  FILE=Rocky-8-GenericCloud.latest.x86_64
  EXT=qcow2
  URL=https://download.rockylinux.org/pub/rocky/8/images/x86_64
  INSTALL="--install mc,tmux,qemu-guest-agent"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_rocky-8.sh"
}

function rocky_9 () {  ### Rocky 9
  VMID=2999
  DESC='Rocky Linux 9 (Blue Onyx)'
  FILE=Rocky-9-GenericCloud.latest.x86_64
  EXT=qcow2
  URL=https://download.rockylinux.org/pub/rocky/9/images/x86_64
  INSTALL="--install mc,tmux,qemu-guest-agent"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_rocky-9.sh"
}

function debian_10 () {  ### Debian 10
  VMID=1010
  DESC='Debian 10 (Buster)'
  FILE=debian-10-generic-amd64
  EXT=qcow2
  URL=https://cloud.debian.org/images/cloud/buster/latest
  INSTALL="--install qemu-guest-agent"
  FIRSTBOOT_INSTALL="--firstboot-install ansible,colordiff,git,haveged,mc,needrestart,tmux,qemu-guest-agent"
  UNINSTALL="--uninstall byobu"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_debian-10.sh"
}

function debian_11 () {  ### Debian 11
  VMID=1111
  DESC='Debian 11 (Bullseye)'
  FILE=debian-11-generic-amd64
  EXT=qcow2
  URL=https://cloud.debian.org/images/cloud/bullseye/latest
  INSTALL="--install qemu-guest-agent"
  FIRSTBOOT_INSTALL="--firstboot-install ansible,colordiff,git,haveged,mc,needrestart,tmux,qemu-guest-agent"
  UNINSTALL="--uninstall byobu"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_debian-11.sh"
}

function debian_12 () {  ### Debian 12
  VMID=1212
  DESC='Debian 12 (Bookworm)'
  FILE=debian-12-generic-amd64
  EXT=qcow2
  URL=https://cloud.debian.org/images/cloud/bookworm/latest
  INSTALL="--install qemu-guest-agent"
  FIRSTBOOT_INSTALL="--firstboot-install ansible,colordiff,git,haveged,mc,needrestart,tmux,qemu-guest-agent"
  UNINSTALL="--uninstall byobu"
  FIRSTBOOT_SCRIPT="--firstboot ./firstboot_debian-12.sh"
}

function run_as_root () {  ### check if script is runnig as root user
  clear
  if [ "$USER" != "root" ]
  then
      echo -e "${red}Please run this script as root.${clear}"
      echo exit 9 ; exit 9 ; exit 9
  fi
}

function linux_selection () {  ### linux selection; ### which linux distribution and version should be used to create a template?
  while true; do
    echo ''
    echo Create an Linux template based on...
    echo '  - [F]  Ubuntu Focal Fossa (20.04 LTS)'
    echo '  - [J]  Ubuntu Jammy Jellyfish (22.04 LTS)'
    echo '  - [0]  Debian 10 (Buster)'
    echo '  - [1]  Debian 11 (Bullseye)'
    echo '  - [2]  Debian 12 (Bookworm)'
    echo '  - [A8] AlmaLinux 8 (Green Obsidian)'
    echo '  - [A9] AlmaLinux 9 (Blue Onyx)'
    echo '  - [R8] Rocky Linux 8 (Green Obsidian)'
    echo '  - [R9] Rocky Linux 9 (Blue Onyx)'
    read -p 'From which Linux distribution and version would you like to create a template? ' OPTION
    case $OPTION in
      [Ff]*     ) LINUX=ubuntu-focal;   break;;
      [Jj]*     ) LINUX=ubuntu-jammy;   break;;
      [0]*      ) LINUX=debian-10;      break;;
      [1]*      ) LINUX=debian-11;      break;;
      [2]*      ) LINUX=debian-12;      break;;
      [Aa][8]*  ) LINUX=alma-8;         break;;
      [Aa][9]*  ) LINUX=alma-9;         break;;
      [Rr][8]*  ) LINUX=rocky-8;        break;;
      [Rr][9]*  ) LINUX=rocky-9;        break;;
      * ) echo Please only answer F, J, 0, 1, 2, A8, A9, R8 or R9 !;;
    esac
  done
}

function storage_selection () {  ### storage selection; ### On which storage, should the template be created on? ### needs to be fixed to only list available bridges and use these as options
  while true; do
    echo ''
    echo Create the template on this storage...
    echo '  - [1] ceph_fs'
    echo '  - [2] raid6'
    echo '  - [3] local'
    echo '  - [4] local-zfs'
    echo '  - [5] local-lvm'
    read -p 'On which storage would you like to create a template on? ' NUM
    case $NUM in
      [1]* ) STORAGE=ceph_ssd;  break;;
      [2]* ) STORAGE=raid6;     break;;
      [3]* ) STORAGE=local;     break;;
      [4]* ) STORAGE=local-zfs; break;;
      [5]* ) STORAGE=local-lvm; break;;
      * ) echo Please only answer 1, 2, 3, 4 or 5 !;;
    esac
  done
}

function network_selection () {  ### network selection; ### needs to be fixed to only list available bridges and use these as options
  while true; do
    echo ''
    echo Which network bridge should be used...
    echo 'Network bridge vmbr0 [0]'
    echo 'Network bridge vmbr1 [1]'
    read -p 'Choose network bridge (default is vmbr0)? ' VMBR
    case $VMBR in
      [0]* ) BRIDGE=vmbr0; break;;
      [1]* ) BRIDGE=vmbr1; break;;
      * )    BRIDGE=vmbr0; break;;
    esac
  done
}

function keyboard_selection () {  ### keyboard selection; ### available keyboard options is listed here https://pve.proxmox.com/pve-docs/qm.1.html
  while true; do
    echo ''
    echo Which keyboard layout to use.
    echo '  - [1] sv    (default)'
    echo '  - [2] en-us'
    echo '  - [3] en-gb'
    read -p 'Select keyboard layout ? ' KB
    case "$KB" in
      [1]* ) KEYBOARD=sv;     break;;
      [2]* ) KEYBOARD=en-us;  break;;
      [3]* ) KEYBOARD=en-gb;  break;;
      * ) KEYBOARD=sv;     break;;
    esac
  done
}

function pool_selection () {  ### pool selection
  while true; do
    echo ''
    echo 'Do you want to add the vm/template to a pool (called "tmpl")...'
    read -p 'Add vm/template to a pool? [Y]es (default) or [N]o.' TAG
    case $TAG in
      [Yy]* ) POOL="--pool tmpl"; break;;
      [Nn]* ) POOL=; break;;
      * ) POOL="--pool tmpl"; break;;
    esac
  done
}

function temporary_image () {  ### temporary image file
  echo ''
  echo Copy \'$LINUX-downloaded.$EXT\' to a temporary image file
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo Command: cp $LINUX-downloaded.$EXT  $LINUX-modified.$EXT
    cp $LINUX-downloaded.$EXT $LINUX-modified.$EXT
  else
    cp $LINUX-downloaded.$EXT $LINUX-modified.$EXT > /dev/null
  fi
  printf " ${green}\xE2\x9C\x94${clear} $LINUX-modified.$EXT has been saved.\n"
}

function customize_image () {  ### customize image file
  echo ''
  echo Customize \'$LINUX-modified.$EXT\' image file
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo Command: virt-customize --add $LINUX-modified.$EXT $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_COMMAND $FIRSTBOOT_SCRIPT
    virt-customize --add $LINUX-modified.$EXT $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_COMMAND $FIRSTBOOT_SCRIPT
  else
    virt-customize --add $LINUX-modified.$EXT $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_COMMAND $FIRSTBOOT_SCRIPT > /dev/null
  fi
  printf " ${green}\xE2\x9C\x94${clear} The '"$LINUX-modified.$EXT"' image file has been customized.\n"
}

function sysprep () {  ### sysprep image file
  echo ''
  echo Sysprep the \'$LINUX-modified.$EXT\' image file
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo Command: virt-sysprep --add $LINUX-modified.$EXT
    virt-sysprep --add $LINUX-modified.$EXT
  else
    virt-sysprep --add $LINUX-modified.$EXT > /dev/null
  fi
  printf " ${green}\xE2\x9C\x94${clear} The '"$LINUX-modified.$EXT"' image file has been sysprepped.\n"
}

function remove_machine_id () {  ### remove machine-id from image file
  echo ''
  echo Remove machine-id from the \"$LINUX-modified.$EXT\" image file
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo Command: virt-customize --add $LINUX-modified.$EXT --truncate /etc/machine-id
    virt-customize --add $LINUX-modified.$EXT --truncate /etc/machine-id
  else
    virt-customize --add $LINUX-modified.$EXT --truncate /etc/machine-id > /dev/null
  fi
  printf " ${green}\xE2\x9C\x94${clear} Machine-id has been removed from the '"$LINUX-modified.$EXT"' image file.\n"
}

function create_vm () {  ### create virtual machine (vm)
  echo ''
  echo Create Virtual Machine
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm create $VMID --name $LINUX-cloudinit-template --memory $MEMORY --cores $CORES --sockets $SOCKETS --cpu $CPU --net0 virtio,bridge=$BRIDGE $POOL
    qm create $VMID --name $LINUX-cloudinit-template --memory $MEMORY --cores $CORES --sockets $SOCKETS --cpu $CPU --net0 virtio,bridge=$BRIDGE $POOL
  else
    qm create $VMID --name $LINUX-cloudinit-template --memory $MEMORY --cores $CORES --sockets $SOCKETS --cpu $CPU --net0 virtio,bridge=$BRIDGE $POOL > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} VM created\n"
  printf "       * Template: $LINUX-cloudinit-template\n"
  printf "       * VMID:     $VMID\n"
  printf "       * CPU type: $CPU\n"
  printf "       * Sockets:  $SOCKETS\n"
  printf "       * Cores:    $CORES\n"
  printf "       * Memory:   $MEMORY\n"
  printf "       * Bridge:   $BRIDGE\n"
  printf "       * Pool:     $POOL\n"
}

function import_disk () {  ### import disk image as system disk to vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm disk import $VMID $LINUX-modified.$EXT $STORAGE
    qm disk import $VMID $LINUX-modified.$EXT $STORAGE
  else
    qm disk import $VMID $LINUX-modified.$EXT $STORAGE > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Disk image imported\n"
  printf "       * Storage:  $STORAGE\n"
}

function system_disk () {  ### configure system disk on vm and configure system disk as boot disk on vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --scsihw virtio-scsi-pci --scsi0 $STORAGE:vm-$VMID-disk-0,discard=on,ssd=1
    qm set $VMID --scsihw virtio-scsi-pci --scsi0 $STORAGE:vm-$VMID-disk-0,discard=on,ssd=1
  else
    qm set $VMID --scsihw virtio-scsi-pci --scsi0 $STORAGE:vm-$VMID-disk-0,discard=on,ssd=1 > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Configured boot disk\n"

  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --boot c --bootdisk scsi0
    qm set $VMID --boot c --bootdisk scsi0
  else
    qm set $VMID --boot c --bootdisk scsi0 > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Configured system disk to boot disk\n"
}

function create_cloud_init () {  ### create cloud-init drive on vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --ide2 $STORAGE:cloudinit
    qm set $VMID --ide2 $STORAGE:cloudinit
  else
    qm set $VMID --ide2 $STORAGE:cloudinit > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Cloud-init drive created\n"
}

function serial_vga () {  ### configure serial terminal on vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --serial0 socket --vga serial0
    qm set $VMID --serial0 socket --vga serial0
  else
    qm set $VMID --serial0 socket --vga serial0 > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Serial VGA configured\n"
}

function enable_agent () {  ### enable agent on vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --agent enabled=1
    qm set $VMID --agent enabled=1
  else
    qm set $VMID --agent enabled=1 > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Agent enable\n"
}

function configure_network () {  ### configure network setting on vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --ipconfig0 ip=dhcp
    qm set $VMID --ipconfig0 ip=dhcp
  else
    qm set $VMID --ipconfig0 ip=dhcp > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} DHCP configured\n"
}

function configure_keyboard () {  ### configure keyboard on vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --keyboard $KEYBOARD
    qm set $VMID --keyboard $KEYBOARD
  else
    qm set $VMID --keyboard $KEYBOARD > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Keyboard configured\n"
  printf "       * Keyboard: $KEYBOARD\n"
}

function add_description () {  ### add description to vm
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm set $VMID --description "Server: <br> OS: $DESC <br> Created: $START_TIME"
    qm set $VMID --description "Server: <br> OS: $DESC <br> Created: $START_TIME"
  else
    qm set $VMID --description "Server: <br> OS: $DESC <br> Created: $START_TIME" > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Added Description\n"
  printf "       * OS:       $DESC\n"
  printf "       * Created:  $START_TIME\n"
}

function convert_to_template () {  ### convert image to template
  if [ $TEST = true ]; then
    echo ''
    echo Command: qm template $VMID
    qm template $VMID
  else
    qm template $VMID > /dev/null
  fi
  printf "   ${green}\xE2\x9C\x94${clear} Converted to template\n"
}

function download_source () {  ### download source image file
  echo ''
  echo Download source image file
  echo '############################################################################'
  if [ ! -f $LINUX-downloaded.$EXT ]; then
    if [ $TEST = true ]; then
      echo Command: wget $URL/$FILE.$EXT --progress=bar --show-progress -O $LINUX-downloaded.$EXT
      wget $URL/$FILE.$EXT --progress=bar --show-progress -O $LINUX-downloaded.$EXT
    else
      wget $URL/$FILE.$EXT --quiet --no-verbose  --progress=bar --show-progress -O $LINUX-downloaded.$EXT > /dev/null
    fi
    printf " ${green}\xE2\x9C\x94${clear} The file '"$FILE.$EXT"' downloaded.\n"
  else
    if [ $KEEP = true ]; then
      printf " \xE2\x9C\x94 The file download skipped (KEEP=true).\n"
    else
      if [ $TEST = true ]; then
        echo Command: wget $URL/$FILE.$EXT --progress=bar --show-progress -O $LINUX-downloaded.$EXT
        wget $URL/$FILE.$EXT --progress=bar --show-progress -O $LINUX-downloaded.$EXT
      else
        wget $URL/$FILE.$EXT --quiet --no-verbose  --progress=bar --show-progress -O $LINUX-downloaded.$EXT > /dev/null
      fi
      printf " ${green}\xE2\x9C\x94${clear} The file '"$FILE.$EXT"' downloaded.\n"
    fi
  fi
}

function remove_source () {  ### remove existing source image
  if test -f "$LINUX-modified.$EXT"; then
    echo ''
    echo Remove modified image file
    echo '############################################################################'
    if [ $TEST = true ]; then
      echo Command: rm $LINUX-modified.$EXT
      rm $LINUX-modified.$EXT
    else
      rm $LINUX-modified.$EXT > /dev/null
    fi
    printf " ${green}\xE2\x9C\x94${clear} '"$LINUX-modified.$EXT"' image file removed.\n"
  fi

  if test -f "$LINUX-downloaded.$EXT"; then
    echo ''
    echo Remove downloaded image file
    echo '############################################################################'
    if [ $KEEP = true ]; then
      printf " \xE2\x9C\x94 Keeping '"$LINUX-downloaded.$EXT"' image file.\n"
    else
      if [ $TEST = true ]; then
        echo Command: rm $LINUX-downloaded.$EXT
        rm $LINUX-downloaded.$EXT
      else
        rm $LINUX-downloaded.$EXT > /dev/null
      fi
      printf " ${green}\xE2\x9C\x94${clear} '"$LINUX-downloaded.$EXT"' image file removed.\n"
    fi
  fi
}

function destroy_vm () {  ### destroy and purge existing template
  echo ''
  echo Purge Template $VMID from related configurations.
  echo '############################################################################'
  if [ $TEST = true ]; then
    echo Command: qm destroy $VMID --destroy-unreferenced-disks 1 --purge 1
    qm destroy $VMID --destroy-unreferenced-disks 1 --purge 1
  else
    qm destroy $VMID --destroy-unreferenced-disks 1 --purge 1 > /dev/null
  fi
  printf " ${green}\xE2\x9C\x94${clear} Template "$VMID" purged.\n"
}

function install_requirement () {  ### install required packages (libguestfs-tools) if needed (only required once)
  if ! [ -x "$(command -v virt-customize)" ]; then
    echo ''
    echo Install package libguestfs-tools since it\'s not installed and is required.
    echo '###########################################################################'
    if [ $TEST = true ]; then
      echo Command: apt update
      apt update
      echo Command: apt install libguestfs-tools -y
      apt install libguestfs-tools -y
    else
      apt update > /dev/null
      apt install libguestfs-tools -y > /dev/null
    fi
    printf " ${green}\xE2\x9C\x94${clear} Package 'libguestfs-tools' installed.\n"
  fi
}

function start_time () {  ### start time
  START_TIME=`date +'%F %T %Z'`
  echo ''
  echo ''
  echo Script started at $START_TIME.
}

function stop_time () {  ### end time
  STOP_TIME=`date +'%F %T %Z'`
  echo ''
  echo Template have been created!
  echo The script was ready at: $STOP_TIME.
  echo ''
}

function write_variables () {  ### write out variables if TEST=true
  echo ''
  echo Write out variables
  echo '############################################################################'
  echo USER=$USER
  echo OPTION=$OPTION
  echo LINUX=$LINUX
  echo NUM=$NUM
  echo STORAGE=$STORAGE
  echo VMID=$VMID
  echo DESC=$DESC
  echo FILE=$FILE
  echo EXT=$EXT
  echo KB=$KB
  echo KEYBOARD=$KEYBOARD
  echo URL=$URL
  echo INSTALL=$INSTALL
  echo FIRSTBOOT_INSTALL=$FIRSTBOOT_INSTALL
  echo UNINSTALL=$UNINSTALL
  echo FIRSTBOOT_COMMAND=$FIRSTBOOT_COMMAND
  echo FIRSTBOOT_SCRIPT=$FIRSTBOOT_SCRIPT
  echo virt-customize --add $FILE $INSTALL $FIRSTBOOT_INSTALL $UNINSTALL $FIRSTBOOT_SCRIPT
}

function main () {  ### Main function
  run_as_root

  linux_selection
  storage_selection
  network_selection
  keyboard_selection
  pool_selection

if [ $LINUX == "ubuntu-focal" ];   then ubuntu_focal; fi
if [ $LINUX == "ubuntu-jammy" ];   then ubuntu_jammy; fi
if [ $LINUX == "alma-8" ];         then alma_8; fi
if [ $LINUX == "alma-9" ];         then alma_9; fi
if [ $LINUX == "rocky-8" ];        then rocky_8; fi
if [ $LINUX == "rocky-9" ];        then rocky_9; fi
if [ $LINUX == "debian-10" ];      then debian_10; fi
if [ $LINUX == "debian-11" ];      then debian_11; fi
if [ $LINUX == "debian-12" ];      then debian_12; fi

if [ $TEST = true ]; then write_variables; fi
if [ $SKIP = true ]; then echo exiting...; exit; exit; fi

  start_time
  install_requirement
  remove_source
  destroy_vm
  download_source
  temporary_image
  customize_image
  sysprep
  remove_machine_id
  create_vm
  import_disk
  system_disk
  create_cloud_init
  serial_vga
  enable_agent
  configure_network
  configure_keyboard
  add_description
  convert_to_template
  remove_source
  stop_time

  exit 0 ; exit 0
}

main
